from fnmatch import fnmatchcase
from functools import cached_property
from typing import Any, Optional, Sequence

from pydantic import BaseModel, ConfigDict, Field, TypeAdapter

from .driver import Driver
from .utils import (
    allow_singular,
    allow_singular_none,
)


class MatchConfig(BaseModel):
    model_config = ConfigDict(
        frozen=True,
        arbitrary_types_allowed=True,
        ignored_types=(cached_property,),
        extra="forbid",
    )

    host: Optional[list[str]] = None
    host_val = allow_singular("host")

    component: list[str] = Field(default_factory=list)
    component_val = allow_singular_none("component")

    driver: Driver

    async def applicable(self) -> bool:
        if self.host is not None:
            host = await self.driver.run_output("hostname")
            if not any(fnmatchcase(host, pattern) for pattern in self.host):
                return False
        return True

    @staticmethod
    async def components(config: Sequence["MatchConfig"]) -> list[str]:
        result = []
        for match in config:
            if await match.applicable():
                result.extend(match.component)
        return result


MatchConfigList = TypeAdapter(list[MatchConfig])


def parse_config(config: Any, *, driver: Driver) -> list[MatchConfig]:
    config = TypeAdapter(list[dict]).validate_python(config)
    return MatchConfigList.validate_python(
        [{**match, "driver": driver} for match in config]
    )



import json
from abc import ABC, abstractmethod
from typing import Any, Union

from bs4 import BeautifulSoup
from jsonpath_ng import JSONPath as JSONPathT  # type: ignore
from jsonpath_ng.ext import parse as jsonpath_parse  # type: ignore
from pydantic import BaseModel, ConfigDict, Field, TypeAdapter, field_validator

from .filters import Filters, choose
from .utils import http_get


class ValueC(BaseModel, ABC):
    model_config = ConfigDict(frozen=True, extra="forbid")

    @abstractmethod
    async def compute(self) -> str:
        raise NotImplementedError


async def compute(value: Union[str, ValueC]) -> str:
    if isinstance(value, str):
        return value
    return await value.compute()


class Derived(ValueC, ABC):
    @abstractmethod
    async def derived_value(self, contents: str) -> str:
        raise NotImplementedError

    async def compute(self) -> str:
        # mypy doesn't like inherited pydantic fields with aliases:
        # https://github.com/pydantic/pydantic/issues/11009
        # https://github.com/python/mypy/issues/18216
        # therefore the base field is defined on every subclass
        base = await compute(self.base)  # type: ignore
        return await self.derived_value(base)


class URL(Derived):
    base: "Value" = Field(..., alias="url")

    async def derived_value(self, contents: str) -> str:
        return await http_get(contents)


class JSONPath(Derived, Filters):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    base: "Value"
    jsonpath: JSONPathT

    @field_validator("jsonpath", mode="before")
    @classmethod
    def parse_jsonpath(cls, value):
        return jsonpath_parse(value)

    async def derived_value(self, contents: str) -> str:
        json_contents = json.loads(contents)
        candidates = [
            candidate.value for candidate in self.jsonpath.find(json_contents)
        ]
        return choose(candidates, self.filters())


class Format(Derived):
    base: "Value"
    format: str  # pylint:disable=redefined-builtin

    async def derived_value(self, contents: str) -> str:
        return self.format.format(contents)


class HTMLLinks(Derived, Filters):
    base: "Value" = Field(..., alias="links")

    async def derived_value(self, contents: str) -> str:
        soup = BeautifulSoup(contents, "html.parser")
        candidates: list[str] = list(
            filter(None, (link.get("href") for link in soup.find_all("a")))
        )
        return choose(candidates, self.filters())


Value = Union[str, URL, HTMLLinks, Format, JSONPath]


ValueAdapter: TypeAdapter[Value] = TypeAdapter(Value)


def parse_value(value: Any) -> Value:
    return ValueAdapter.validate_python(value)



import argparse
import sys
from pathlib import Path
from typing import Optional

import trio
from typed_argparse import TypedArgs

from .driver import LocalDriver
from .manager import Manager
from .state import DB, DB_PATH


class Args(TypedArgs):
    pass


def parse_args(args: Optional[list[str]] = None) -> Args:
    result = argparse.ArgumentParser(description="Install the environment")
    return Args.from_argparse(result.parse_args(args))


async def main() -> int:
    parse_args()

    db = DB(DB_PATH)
    driver = LocalDriver()
    manager = Manager(db=db, driver=driver, data_path=Path())

    result = await manager.install()

    if result.installed:
        print(
            f"{len(result.installed)} packages installed or updated: {', '.join(p.name for p in result.installed)}"
        )
    if result.failed:
        for p, exc in result.failed:
            print(f"Failed to install {p.name}: {exc}")

    if not result.installed and not result.failed:
        print("Everything up to date.")

    return len(result.failed)


def sync_main() -> None:
    sys.exit(trio.run(main))



import shlex
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Optional

from .driver import Driver
from .utils import async_cached


class Extractor(ABC):
    def __init__(self, *, driver: Driver):
        self.driver = driver

    async def extract(self, *, archive: Path, target_directory: Path) -> None:
        async with self.driver.tempfile(kind="directory") as tmpdir:
            await self.extract_exact(archive=archive, target_directory=tmpdir)

            source_dir = tmpdir
            stripped = 0

            while True:
                contents = await self.driver.find(source_dir, mindepth=1, maxdepth=1)
                if len(contents) != 1:
                    # Multiple files/directories on this level
                    break

                element = contents[0]
                if not await self.driver.is_dir(element):
                    # The only item is a file
                    break

                source_dir = source_dir / element
                stripped += 1
                if stripped >= 10:
                    raise ValueError(
                        f"Too many nested directories after extracting, got {source_dir}."
                    )

            for element in await self.driver.find(source_dir, mindepth=1, maxdepth=1):
                await self.driver.copy(element, target_directory / element.name)

    @abstractmethod
    async def extract_exact(self, *, archive: Path, target_directory: Path) -> None:
        pass


class Tar(Extractor):
    def __init__(self, *, driver: Driver, extra: Optional[list[str]] = None):
        super().__init__(driver=driver)
        self.extra = extra or []

    @async_cached
    async def tar(self) -> str:
        return await self.driver.find_executable("gtar", "tar")

    async def extract_exact(self, *, archive: Path, target_directory: Path) -> None:
        await self.driver.run(
            await self.tar(),
            "--extract",
            "--directory",
            target_directory,
            *self.extra,
            "--file",
            archive,
        )


class Unzip(Extractor):
    async def extract_exact(self, *, archive: Path, target_directory: Path) -> None:
        await self.driver.run("unzip", "-o", "-qq", archive, "-d", target_directory)


def _guess_extractor(url: str, *, driver: Driver) -> Extractor:
    if url.endswith(".tar"):
        return Tar(driver=driver)
    elif url.endswith(".tar.gz") or url.endswith(".tgz"):
        return Tar(driver=driver, extra=["-z"])
    elif url.endswith(".tar.bz2"):
        return Tar(driver=driver, extra=["-j"])
    elif url.endswith(".tar.xz") or url.endswith(".txz"):
        return Tar(driver=driver, extra=["-J"])
    elif url.endswith(".zip"):
        return Unzip(driver=driver)
    else:
        raise ValueError(f"Unknown archive format: {url}")


async def get_redirect_url(url: str, *, driver: Driver) -> str:
    return await driver.run_output(
        "curl",
        "--write-out",
        "%{url_effective}",
        "--head",
        "--no-include",
        "--silent",
        "--show-error",
        "--location",
        url,
    )


async def get_extractor(url: str, *, driver: Driver) -> Extractor:
    try:
        return _guess_extractor(url, driver=driver)
    except ValueError:
        url = await get_redirect_url(url, driver=driver)
        return _guess_extractor(url, driver=driver)


class RawExtractor(ABC):
    def __init__(self, *, driver: Driver):
        self.driver = driver

    @abstractmethod
    async def extract(self, *, archive: Path, target: Path) -> None:
        pass


class ShellRedirectRawExtractor(RawExtractor):
    def __init__(self, *, driver: Driver, command: str):
        super().__init__(driver=driver)
        self.command = command

    async def extract(self, *, archive: Path, target: Path) -> None:
        await self.driver.run(
            "sh",
            "-c",
            f"{self.command} < {shlex.quote(str(archive))} > {shlex.quote(str(target))}",
        )


class Move(RawExtractor):
    async def extract(self, *, archive: Path, target: Path) -> None:
        await self.driver.copy(archive, target)


def _guess_single_extractor(url: str, *, driver: Driver) -> RawExtractor:
    if url.endswith(".gz"):
        return ShellRedirectRawExtractor(driver=driver, command="gunzip")
    elif url.endswith(".xz"):
        return ShellRedirectRawExtractor(driver=driver, command="xzcat")
    elif url.endswith(".bz2"):
        return ShellRedirectRawExtractor(driver=driver, command="bunzip2")
    else:
        raise ValueError(f"Unknown archive format: {url}")


async def get_single_extractor(url: str, *, driver: Driver) -> RawExtractor:
    try:
        return _guess_single_extractor(url, driver=driver)
    except ValueError:
        url = await get_redirect_url(url, driver=driver)
        try:
            return _guess_single_extractor(url, driver=driver)
        except ValueError:
            return Move(driver=driver)
